@@ -161,6 +161,18 @@ int libssl_ticket_key_callback(SSL *s, OpenSSLTLSTicketKeysRing& keyring, unsign
   return 1;
 }
 
+static long libssl_server_name_callback(SSL* ssl, int* al, void* arg)
+{
+  (void) al;
+  (void) arg;
+
+  if (SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name)) {
+    return SSL_TLSEXT_ERR_OK;
+  }
+
+  return SSL_TLSEXT_ERR_NOACK;
+}
+
 static void libssl_info_callback(const SSL *ssl, int where, int ret)
 {
   SSL_CTX* sslCtx = SSL_get_SSL_CTX(ssl);
@@ -566,7 +578,7 @@ OpenSSLTLSTicketKey::OpenSSLTLSTicketKey()
 #endif /* HAVE_LIBSODIUM */
 }
 
-OpenSSLTLSTicketKey::OpenSSLTLSTicketKey(ifstream& file)
+OpenSSLTLSTicketKey::OpenSSLTLSTicketKey(std::ifstream& file)
 {
   file.read(reinterpret_cast<char*>(d_name), sizeof(d_name));
   file.read(reinterpret_cast<char*>(d_cipherKey), sizeof(d_cipherKey));
@@ -661,6 +673,17 @@ std::unique_ptr<SSL_CTX, void(*)(SSL_CTX*)> libssl_init_server_context(const TLS
 
   if (config.d_preferServerCiphers) {
     sslOptions |= SSL_OP_CIPHER_SERVER_PREFERENCE;
+#ifdef SSL_OP_PRIORITIZE_CHACHA
+    sslOptions |= SSL_OP_PRIORITIZE_CHACHA;
+#endif /* SSL_OP_PRIORITIZE_CHACHA */
+  }
+
+  if (!config.d_enableRenegotiation) {
+#ifdef SSL_OP_NO_RENEGOTIATION
+    sslOptions |= SSL_OP_NO_RENEGOTIATION;
+#elif defined(SSL_OP_NO_CLIENT_RENEGOTIATION)
+    sslOptions |= SSL_OP_NO_CLIENT_RENEGOTIATION;
+#endif
   }
 
   SSL_CTX_set_options(ctx.get(), sslOptions);
@@ -682,6 +705,17 @@ std::unique_ptr<SSL_CTX, void(*)(SSL_CTX*)> libssl_init_server_context(const TLS
     SSL_CTX_sess_set_cache_size(ctx.get(), config.d_maxStoredSessions);
   }
 
+#ifdef SSL_MODE_RELEASE_BUFFERS
+  if (config.d_releaseBuffers) {
+    SSL_CTX_set_mode(ctx.get(), SSL_MODE_RELEASE_BUFFERS);
+  }
+#endif
+
+  /* we need to set this callback to acknowledge the server name sent by the client,
+     otherwise it will not stored in the session and will not be accessible when the
+     session is resumed, causing SSL_get_servername to return nullptr */
+  SSL_CTX_set_tlsext_servername_callback(ctx.get(), &libssl_server_name_callback);
+
   std::vector<int> keyTypes;
   /* load certificate and private key */
   for (const auto& pair : config.d_certKeyPairs) {
@@ -762,4 +796,19 @@ std::unique_ptr<FILE, int(*)(FILE*)> libssl_set_key_log_file(std::unique_ptr<SSL
 #endif /* HAVE_SSL_CTX_SET_KEYLOG_CALLBACK */
 }
 
+std::string libssl_get_error_string()
+{
+  BIO *mem = BIO_new(BIO_s_mem());
+  ERR_print_errors(mem);
+  char *p;
+  size_t len = BIO_get_mem_data(mem, &p);
+  std::string msg(p, len);
+  // replace newlines by /
+  if (msg.back() == '\n') {
+    msg.pop_back();
+  }
+  std::replace(msg.begin(), msg.end(), '\n', '/');
+  BIO_free(mem);
+  return msg;
+}
 #endif /* HAVE_LIBSSL */
